#if !defined(INOUT_INCLUDED)
#define INOUT_INCLUDED

#include <cstring>

//#include "config.h"

/************** file name operations ****************/

typedef const char *const_string;

// remove the extension 'ext' from 'fullname' if present
// allocate a new string
// string 'ext' must contain the leading period
char *extractbasename (const_string fullname, const_string ext) {
	unsigned int fulllen = strlen(fullname);
	if (fulllen>4 && strcmp(fullname+fulllen-4,ext)==0) {
		char *basename = new char[fulllen-4+1];
		memcpy (basename, fullname, fulllen-4);
		basename[fulllen-4] = 0;
		return basename;
	} else {
		char *basename = new char[fulllen+1];
		strcpy (basename, fullname);
		return basename;
	}
}

bool is_suffix (const_string str, const_string suffix) {
	long    len = strlen(str);
	long suflen = strlen(suffix);
	return (len>suflen) && (strcmp(str+len-suflen,suffix)==0);
}




/************** argument parsing ****************/

bool isstringempty(const char *str) {
	return (str==NULL) || (str[0]==0);
}

bool argtol(const_string str, long &l) {
	char *end;
	l = strtol(str, &end,10);
	return *end == 0;
}

long opttol(const_string val, char key) {
	long l;
	if (!argtol(val,l)) {
		printf("<%s> was passed as value for the option -%c but is no integer.\n", val, key);
		exit(1);
	}
	return l;
}

bool argtof(const_string str, float &f) {
	char *end;
	f = strtod(str, &end);
	return *end == 0;
}

float opttof(const_string val, char key) {
	float f;
	if (!argtof(val,f)) {
		printf("<%s> was passed as value for the option -%c but is no float number.\n", val, key);
		exit(1);
	}
	return f;
}



/* deallocation safe wrappers of commonly used types */

// preserves deallocation of the string if a scope is left
// Never copy objects of this class!
struct safe_string {
	char *str;
	public:
	// string length counted without terminating zero
	safe_string(long len) : str(new char[len+1]) {clear();};
	safe_string(char *newstr) : str(newstr) {};
	~safe_string() {delete[] str;};

	inline long len() const {return strlen(str);};
	inline void clear() {str[0] = 0;};
	inline void append(const_string appstr) {strcat(str,appstr);};
	inline const_string cstr() const {return str;};
};

struct safe_signal {
	float *f;
	public:
	safe_signal() : f(NULL) 
	{
	};
	safe_signal(float *newf) : f(newf) 
	{
	};
	safe_signal(long len) : f(new float[len]) {};
	~safe_signal() {if (f!=NULL) {delete[] f;}};

	void set (long len, float value) {
		float *data = f;
		for (;len>0;len--) {
			*data = value;
			data++;
		}
	};
	inline void replace (float *by) {
		if (f!=NULL) {
			delete[] f;
		}
		f = by;
	};
	inline void swap(safe_signal &with) {
		float *c;
		c      = f;
		f      = with.f;
		with.f = c;
	}
	inline float *cptr() 
	{
		return f;
	};
	inline float *untied_cptr() {
		float *f0 = f;
		f = NULL;
		return f0;
	};
};


/************** execution of external programs ****************/
void startprog(const_string prog, const_string filename, const_string extension, bool asynchronous = false){
	//printf("startprog(\"%s\",\"%s\",\"%s\",%d)\n", prog, filename, extension, asynchronous);
	long len = strlen(prog)+1+strlen(filename)+strlen(extension)+1;
	if (asynchronous) {
		len+=2;
	}
	safe_string prompt(len);
	prompt.append(prog);
	prompt.append(" ");
	prompt.append(filename);
	prompt.append(extension);
	if (asynchronous) {
		prompt.append(" &");
	}
	system(prompt.cstr());
}


/****************** more specialized fopen() and fclose() routines ************/
struct safe_file {
	typedef char text_line[256];
	FILE *file;
	public:
	safe_file(FILE *newfile) : file(newfile) {};
	~safe_file() {fclose(file);};
	FILE *cfile() const {return file;};
	bool readline (text_line &ln);
};


bool safe_file::readline (text_line &ln) {
	return fgets (&ln[0], sizeof(text_line), file) != NULL;
}

FILE *wfopen(const char *filename, const char *extension, const char *mode) {
	//printf("wfopen(\"%s\",\"%s\",\"%s\")\n", filename, extension, mode);
	FILE *file;
	safe_string fullname(strlen(filename)+strlen(extension)+1);
	fullname.append(filename);
	fullname.append(extension);
	file = fopen(fullname.cstr(),mode);
	if(file==NULL){
		printf("file <%s> could not opened with mode '%s'\n", fullname.cstr(), mode);
		exit(1);  // dangerous, because other resources can't be deallocated
	}
	return file;
}

// can work properly only if you call it _after_ you closed the image file
// that means, that you have to place the call _outside_ of the visibility scope of the safe_file handle
void view_result(const char *filename, const char *extension, bool graphic, const char *viewer){
	if(graphic) {
		startprog (viewer, filename, extension, true);
	}
}


void printhbar() {
	printf("\n----------------------------\n");
}


/************** some arithmetic, a bit misplaced in this file :-) ****************/

// extends the bounds xmin,xmax so that x is within them
inline void refresh_bounds (float x, float &xmin, float &xmax) {
	if(xmin>=x) {xmin=x;}
	if(xmax<=x) {xmax=x;}
}

#ifndef HAVE_FMAX
inline float fmax (float x, float y) {
	if (x>=y) {
		return x;
	} else {
		return y;
	}
}
#endif

#ifndef HAVE_FMIN
inline float fmin (float x, float y) {
	if (x<=y) {
		return x;
	} else {
		return y;
	}
}
#endif

inline int sgn (float x) {
	if (x>0) {
		return 1;
	} else if (x<0) {
		return -1;
	} else {
		return 0;
	}
}


/************** help for debugging ****************/

void write_signal (const_string filename, const float *data, long len) {
/*	safe_file file(fopen(filename,"w"));
	long i;
	for(i=0; i<len; i++) {
		fprintf(file.cfile(), "%6ld: %f\n", i, data[i]);
	}
	*/
};


#endif //INOUT_INCLUDED
