// wavelets.h

/*
 * Classes for calculating the wavelet transform.
 */
#ifndef _WAVELETS_H_
#define _WAVELETS_H_

//#include "../audiosystem.h"
#include "filtercoeffs.h"

#define SAMPLE DOUBLE

typedef int (*t_ForwardFunction) 
	(SAMPLE* input, int inCount, SAMPLE* details, SAMPLE* approx,
		float* lowFilter, float* highFilter, int filterLength, SAMPLE* history);

typedef void (*t_InverseFunction) 
	(SAMPLE* output, SAMPLE* details, SAMPLE* approx, int coeffCount,
		int outCount, float* lowFilter, float* highFilter, int filterLength);

/*
 * Manages a set of wavelet coefficients in different levels/scales.
 */
class WaveletCoeffs {
public:
	WaveletCoeffs();
	~WaveletCoeffs();
	void reset(int totalSignalLength, int filterLength);
	// coeffs must point somewhere inside CoeffBuffer !
	void addLevel(SAMPLE* coeffs, int count);
	// coeffs must point somewhere inside CoeffBuffer !
	void setApprox(SAMPLE* coeffs, int count);
	int getLevelCount();
	SAMPLE* getDetails(int level, int* count);
	int getDetailCount(int level);
	SAMPLE* getApprox(int* count);
	int getTotalSignalLength() { return m_totalSignalLength; }
	SAMPLE* getCoeffBuffer() { return m_coeffBuffer; }
	int getCoeffBufferSize() { return m_coeffBufferSize; }
	void copyFrom(WaveletCoeffs* otherCoeffs);
	
	// linear memory optimization: optional
	bool isLinearMemory() { return m_linMem!=NULL; }
	void setLinearMemory(SAMPLE* lin, int count) { m_linMem=lin; m_linMemCount=count; }
	SAMPLE* getLinearMemory() { return m_linMem; }
	int getLinearMemoryCount() { return m_linMemCount; }
private:
	void internalReset(int totalSignalLength, int bufferSize);
	CArray<SAMPLE*,SAMPLE*> m_coeffs;
	CArray<INT,INT> m_counts;
	SAMPLE* m_approx;
	int m_approxCount;
	int m_totalSignalLength;
	SAMPLE* m_linMem;
	int m_linMemCount;
	SAMPLE* m_coeffBuffer;
	int m_coeffBufferSize;
};

// the history stores the last (filterlength-2) values of
// - the signal: in level=0
// - the previous wavelet domain coefficients: in remaining levels
// the samples in the history are flipped, i.e. history[0]==last sample, history[1]==before last sample
class WaveletHistory {
public:
	WaveletHistory();
	~WaveletHistory();
	void reset(); // sets all history values to 0
	void update(WaveletCoeffs* coeffs, int filterLength);
	void updateSignal(SAMPLE* signal, int signalLength, int filterLength);
	int getLevelCount() { return m_currLevelCount; }
	SAMPLE* getHistory(int level);
	SAMPLE* getHistoryNoCheck(int level);
private:
	void reset(int level);
	void moveSamples(int level, SAMPLE* coeffs, int coeffCount, int count);
	void destroyMemory();
	void assertLevels(int levels);
	CArray<SAMPLE*,SAMPLE*> m_hcoeffs;
	CArray<INT*,INT*> m_hcounts;
	int m_currLevelCount;
};

typedef struct {
	int ncof,ioff,joff;
	float *cc,*cr;
} wavefilt;

/*
 * Provides methods for calculating forward and inverse wavelet transform.
 */
class WaveletTransform {
public:
	WaveletTransform();
	~WaveletTransform();
	void forwardTransform(SAMPLE* input, int inputLength, WaveletCoeffs* coeffs, WaveletHistory* history);
	void inverseTransform(WaveletCoeffs* input, SAMPLE* output);
	int getWaveletCount();
	void setWavelet(int number);
	int getWavelet() { return m_currWavelet; }
	char* getWaveletName();
	int getAlgorithmCount();
	void setAlgorithm(int number);
	int getAlgorithm() { return m_algorithm; }
	char* getAlgorithmName();
private:
	void numericalRecipesForwardTransform(SAMPLE* input, int inputLength, WaveletCoeffs* coeffs);
	void numericalRecipesInverseTransform(WaveletCoeffs* input, SAMPLE* output);
	void generalForwardTransform(SAMPLE* input, int inputLength, WaveletCoeffs* coeffs, WaveletHistory* history);
	void generalInverseTransform(WaveletCoeffs* input, SAMPLE* output);
	void assertWorkingBuffer(int totalSignalLength);
	void delWorkingBuffer();
	void setTransformFuncs();
	int m_currWavelet;
	int m_algorithm;
	SAMPLE* m_workingBuffer;
	int m_workingBufferSize;
	wavefilt m_wfilt;
	t_ForwardFunction m_forwardFunc;
	t_InverseFunction m_inverseFunc;
};


#endif